package de.meinefirma.meinprojekt.crypto;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.security.Security;
import java.util.Date;
import java.util.Iterator;

import org.bouncycastle.bcpg.ArmoredOutputStream;
import org.bouncycastle.bcpg.CompressionAlgorithmTags;
import org.bouncycastle.bcpg.HashAlgorithmTags;
import org.bouncycastle.bcpg.PublicKeyAlgorithmTags;
import org.bouncycastle.bcpg.SymmetricKeyAlgorithmTags;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.openpgp.PGPCompressedData;
import org.bouncycastle.openpgp.PGPCompressedDataGenerator;
import org.bouncycastle.openpgp.PGPEncryptedDataGenerator;
import org.bouncycastle.openpgp.PGPEncryptedDataList;
import org.bouncycastle.openpgp.PGPException;
import org.bouncycastle.openpgp.PGPKeyPair;
import org.bouncycastle.openpgp.PGPLiteralData;
import org.bouncycastle.openpgp.PGPLiteralDataGenerator;
import org.bouncycastle.openpgp.PGPObjectFactory;
import org.bouncycastle.openpgp.PGPOnePassSignatureList;
import org.bouncycastle.openpgp.PGPPrivateKey;
import org.bouncycastle.openpgp.PGPPublicKey;
import org.bouncycastle.openpgp.PGPPublicKeyEncryptedData;
import org.bouncycastle.openpgp.PGPPublicKeyRing;
import org.bouncycastle.openpgp.PGPPublicKeyRingCollection;
import org.bouncycastle.openpgp.PGPSecretKey;
import org.bouncycastle.openpgp.PGPSecretKeyRingCollection;
import org.bouncycastle.openpgp.PGPSignature;
import org.bouncycastle.openpgp.PGPUtil;
import org.bouncycastle.openpgp.operator.PGPDigestCalculator;
import org.bouncycastle.openpgp.operator.jcajce.JcaPGPContentSignerBuilder;
import org.bouncycastle.openpgp.operator.jcajce.JcaPGPDigestCalculatorProviderBuilder;
import org.bouncycastle.openpgp.operator.jcajce.JcaPGPKeyConverter;
import org.bouncycastle.openpgp.operator.jcajce.JcePBESecretKeyDecryptorBuilder;
import org.bouncycastle.openpgp.operator.jcajce.JcePBESecretKeyEncryptorBuilder;
import org.bouncycastle.openpgp.operator.jcajce.JcePGPDataEncryptorBuilder;
import org.bouncycastle.openpgp.operator.jcajce.JcePublicKeyDataDecryptorFactoryBuilder;
import org.bouncycastle.openpgp.operator.jcajce.JcePublicKeyKeyEncryptionMethodGenerator;
import org.bouncycastle.util.io.Streams;

public class OpenPgpBouncyCastleStreamUtil
{
   public static final String ASYMMETRIC_ALGORITHM        = "RSA";
   public static final int    PUBLIC_KEY_ALGORITHM_TAG    = PublicKeyAlgorithmTags.RSA_GENERAL;
   public static final int    SYMMETRIC_KEY_ALGORITHM_TAG = SymmetricKeyAlgorithmTags.AES_128;

   private PGPPublicKeyEncryptedData pbe;

   /** Diese Methode vor allen anderen Methoden einmal aufrufen.
    */
   public static void addBouncyCastleProvider()
   {
      Security.addProvider( new BouncyCastleProvider() );
   }

   /** RSA-Schluesselpaar erzeugen.
    * @param privateKeyFile Dateiname fuer Private Key
    * @param publicKeyFile  Dateiname fuer Public Key
    * @param keyIdName      Schluessel-ID, z.B.: "Mein Name (Test) &lt;ich@myserver.de&gt;"
    * @param passphrase     Passphrase, z.B.: "Meine geheime Passphrase 42"
    * @param armor          falls true ASCII-Datei, sonst binaer
    */
   public static void generateKeyPairFiles( String privateKeyFile, String publicKeyFile,
         String keyIdName, String passphrase, boolean armor ) throws NoSuchAlgorithmException, IOException
   {
      try( OutputStream privateKeyOutputStream = new FileOutputStream( privateKeyFile );
           OutputStream publicKeyOutputStream  = new FileOutputStream( publicKeyFile  ) ) {
         generateKeyPairStreams( privateKeyOutputStream, publicKeyOutputStream, keyIdName, passphrase, armor );
      }
   }

   /** RSA-Schluesselpaar erzeugen.<br>
    * Die Streams privateKeyOutputStream und publicKeyOutputStream werden mit close() geschlossen.
    * @param privateKeyOutputStream OutputStream fuer Private Key
    * @param publicKeyOutputStream  OutputStream fuer Public Key
    * @param keyIdName  Schluessel-ID, z.B.: "Mein Name (Test) &lt;ich@myserver.de&gt;"
    * @param passphrase Passphrase, z.B.: "Meine geheime Passphrase 42"
    * @param armor      falls true ASCII-Datei, sonst binaer
    */
   public static void generateKeyPairStreams( OutputStream privateKeyOutputStream, OutputStream publicKeyOutputStream,
         String keyIdName, String passphrase, boolean armor ) throws NoSuchAlgorithmException, IOException
   {
      KeyPairGenerator kpg = KeyPairGenerator.getInstance( ASYMMETRIC_ALGORITHM, new BouncyCastleProvider() );
      kpg.initialize( 2048 );
      try {
         exportKeyPair( privateKeyOutputStream, publicKeyOutputStream, kpg.generateKeyPair(), keyIdName, passphrase.toCharArray(), armor );
      } catch( PGPException ex ) {
         throw new CryptoException( ex );
      }
   }

   /** Entschluesseln einer Datei.
    * @param decryptedOutFile Entschluesselte Ergebnisdatei
    * @param encryptedInFile  Verschluesselte Inputdatei
    * @param privateKeyFile   Datei mit Private Key
    * @param passphrase       Passphrase, z.B.: "Meine geheime Passphrase 42"
    * @return                 true falls erfolgreich
    */
   public static boolean decryptFile( String decryptedOutFile, String encryptedInFile, String privateKeyFile, String passphrase ) throws IOException
   {
      return decryptStream( new BufferedOutputStream( new FileOutputStream( decryptedOutFile ) ),
                            new BufferedInputStream(  new FileInputStream(  encryptedInFile ) ),
                            new BufferedInputStream(  new FileInputStream(  privateKeyFile ) ),
                            passphrase.toCharArray() );
   }

   /** Entschluesseln eines Input-Streams (die drei Streams werden mit close() geschlossen).
    * @param decryptedOutputStream OutputStream der entschluesselten Ergebnisdaten
    * @param encryptedInputStream  InputStream der verschluesselten Eingangsdaten
    * @param privateKeyInputStream InputStream zum Private Key
    * @param passphrase            Passphrase, z.B.: "Meine geheime Passphrase 42"
    * @return                      true falls erfolgreich
    */
   public static boolean decryptStream( OutputStream decryptedOutputStream, InputStream encryptedInputStream,
         InputStream privateKeyInputStream, char[] passphrase ) throws IOException
   {
      try {
         InputStream encryptedInpStrm = PGPUtil.getDecoderStream( encryptedInputStream );
         PGPObjectFactory pgpF = new PGPObjectFactory( encryptedInpStrm );
         PGPEncryptedDataList enc;
         Object o = pgpF.nextObject();

         // the first object might be a PGP marker packet
         if( o instanceof PGPEncryptedDataList ) {
            enc = (PGPEncryptedDataList) o;
         } else {
            enc = (PGPEncryptedDataList) pgpF.nextObject();
         }

         // find the secret key
         Iterator<?> it = enc.getEncryptedDataObjects();
         PGPPrivateKey sKey = null;
         PGPPublicKeyEncryptedData pbe = null;
         PGPSecretKeyRingCollection pgpSec = new PGPSecretKeyRingCollection( PGPUtil.getDecoderStream( privateKeyInputStream ) );

         while( sKey == null && it.hasNext() ) {
            pbe = (PGPPublicKeyEncryptedData) it.next();
            sKey = findSecretKey( pgpSec, pbe.getKeyID(), passphrase );
         }

         if( pbe == null || sKey == null ) {
            throw new CryptoException( "secret key for message not found." );
         }

         InputStream clear = pbe.getDataStream( new JcePublicKeyDataDecryptorFactoryBuilder()
                                .setProvider( new BouncyCastleProvider() ).setContentProvider( new BouncyCastleProvider() ).build( sKey ) );
         PGPObjectFactory plainFact = new PGPObjectFactory( clear );
         Object message = plainFact.nextObject();

         if( message instanceof PGPCompressedData ) {
            PGPCompressedData cData = (PGPCompressedData) message;
            PGPObjectFactory pgpFact = new PGPObjectFactory( cData.getDataStream() );
            message = pgpFact.nextObject();
         }

         if( message instanceof PGPLiteralData ) {
            PGPLiteralData ld = (PGPLiteralData) message;
            InputStream unc = ld.getInputStream();
            Streams.pipeAll( unc, decryptedOutputStream );
         } else if( message instanceof PGPOnePassSignatureList ) {
            throw new CryptoException( "encrypted message contains a signed message - not literal data." );
         } else {
            throw new CryptoException( "message is not a simple encrypted file - type unknown." );
         }

         if( pbe.isIntegrityProtected() && !pbe.verify() ) {
            return false;
         }
         return true;
      } catch( PGPException ex ) {
         throw new CryptoException( ex );
      } finally {
         try { encryptedInputStream.close();  } catch( Exception e ) {/*ok*/}
         try { decryptedOutputStream.close(); } catch( Exception e ) {/*ok*/}
         try { privateKeyInputStream.close(); } catch( Exception e ) {/*ok*/}
      }
   }

   /** Entschluesseln eines Input-Streams.
    *  Wichtig: Die drei InputStreams werden nicht innerhalb der Methode geschlossen und muessen ausserhalb mit close() geschlossen werden.
    * @param encryptedInputStream  InputStream der verschluesselten Eingangsdaten
    * @param privateKeyInputStream InputStream zum Private Key
    * @param passphrase            Passphrase, z.B.: "Meine geheime Passphrase 42"
    * @return                      InputStream zu den entschluesselten Ergebnisdaten
    */
   public InputStream decryptStream( InputStream encryptedInputStream, InputStream privateKeyInputStream, char[] passphrase ) throws IOException
   {
      try {
         InputStream encryptedInpStrm = PGPUtil.getDecoderStream( encryptedInputStream );
         PGPObjectFactory pgpF = new PGPObjectFactory( encryptedInpStrm );
         PGPEncryptedDataList enc;
         Object o = pgpF.nextObject();

         // the first object might be a PGP marker packet
         if( o instanceof PGPEncryptedDataList ) {
            enc = (PGPEncryptedDataList) o;
         } else {
            enc = (PGPEncryptedDataList) pgpF.nextObject();
         }
         if( enc == null ) {
            throw new CryptoException( "InputStream probably not encrypted." );
         }

         // find the secret key
         Iterator<?> it = enc.getEncryptedDataObjects();
         PGPPrivateKey sKey = null;
         PGPSecretKeyRingCollection pgpSec = new PGPSecretKeyRingCollection( PGPUtil.getDecoderStream( privateKeyInputStream ) );

         this.pbe = null;
         while( sKey == null && it.hasNext() ) {
            this.pbe = (PGPPublicKeyEncryptedData) it.next();
            sKey = findSecretKey( pgpSec, this.pbe.getKeyID(), passphrase );
         }

         if( this.pbe == null || sKey == null ) {
            throw new CryptoException( "secret key for message not found." );
         }

         InputStream clear = this.pbe.getDataStream( new JcePublicKeyDataDecryptorFactoryBuilder()
                                     .setProvider( new BouncyCastleProvider() ).setContentProvider( new BouncyCastleProvider() ).build( sKey ) );
         PGPObjectFactory plainFact = new PGPObjectFactory( clear );
         Object message = plainFact.nextObject();

         if( message instanceof PGPCompressedData ) {
            PGPCompressedData cData = (PGPCompressedData) message;
            PGPObjectFactory pgpFact = new PGPObjectFactory( cData.getDataStream() );
            message = pgpFact.nextObject();
         }

         if( message instanceof PGPLiteralData ) {
            PGPLiteralData ld = (PGPLiteralData) message;
            return ld.getInputStream();
         } else if( message instanceof PGPOnePassSignatureList ) {
            throw new CryptoException( "encrypted message contains a signed message - not literal data." );
         } else {
            throw new CryptoException( "message is not a simple encrypted file - type unknown." );
         }
      } catch( PGPException ex ) {
         throw new CryptoException( ex );
      }
   }

   /** Pruefe die Integritaet.
    */
   public boolean verifyIntegrityProtected()
   {
      if( this.pbe == null ) { return false; }
      if( !this.pbe.isIntegrityProtected() ) { return true; }
      try {
         return this.pbe.verify();
      } catch( Exception e ) {
         throw new CryptoException( e );
      }
   }

   /** Verschluesseln einer Datei.
    * @param encryptedOutFile   Verschluesselte Ergebnisdatei
    * @param inFile             Zu verschlueselnde Inputdatei
    * @param publicKeyFile      Datei mit Public Key
    * @param armor              falls true ASCII-Datei, sonst binaer
    * @param withIntegrityCheck falls true mit Integritaetsueberpruefung
    */
   public static void encryptFile( String encryptedOutFile, String inFile, String publicKeyFile, boolean armor, boolean withIntegrityCheck ) throws IOException
   {
      File file = new File( inFile );
      try( OutputStream encryptedOutputStream = new FileOutputStream( encryptedOutFile );
           InputStream in = new BufferedInputStream( new FileInputStream( file ) );
           InputStream publicKeyInputStream = new BufferedInputStream( new FileInputStream( publicKeyFile ) ) ) {
         encryptStream( encryptedOutputStream, in,
               file.getName(), file.length(), new Date( file.lastModified() ),
               publicKeyInputStream, armor, withIntegrityCheck );
      }
   }

   /** Verschluesseln eines Input-Streams (die drei Streams werden mit close() geschlossen).
    * @param encryptedOutputStream OutputStream der verschluesselten Ergebnisdaten
    * @param in                    InputStream der zu verschlueselnden Inputdaten
    * @param inFilename            Dateiname, wird bei spaeterer Speicherung in Datei verwendet
    * @param inByteLength          Laenge des InputStreams
    * @param inDataLastModified    Datum der letzten Aenderung der Inputdaten
    * @param publicKeyInputStream  InputStream zum Public Key
    * @param armor                 falls true ASCII-Datei, sonst binaer
    * @param withIntegrityCheck    falls true mit Integritaetsueberpruefung
    */
   public static void encryptStream( OutputStream encryptedOutputStream,
         InputStream in, String inFilename, long inByteLength, Date inDataLastModified,
         InputStream publicKeyInputStream, boolean armor, boolean withIntegrityCheck ) throws IOException
   {
      OutputStream encryptedOutpStrm = encryptedOutputStream;

      try {
         if( armor ) {
            encryptedOutpStrm = new ArmoredOutputStream( encryptedOutpStrm );
         }

         PGPPublicKey encPublicKey = readPublicKey( publicKeyInputStream );

         byte[] bytes = compressFile( CompressionAlgorithmTags.ZIP, in, inFilename, inByteLength, inDataLastModified );

         SecureRandom rand = new SecureRandom();
         PGPEncryptedDataGenerator encGen = new PGPEncryptedDataGenerator(
               new JcePGPDataEncryptorBuilder( SYMMETRIC_KEY_ALGORITHM_TAG ).setWithIntegrityPacket( withIntegrityCheck )
                   .setSecureRandom( rand ).setProvider( new BouncyCastleProvider() ) );
         encGen.addMethod( new JcePublicKeyKeyEncryptionMethodGenerator( encPublicKey ).setProvider( new BouncyCastleProvider() ).setSecureRandom( rand ) );

         OutputStream cOut = encGen.open( encryptedOutpStrm, bytes.length );
         try { cOut.write( bytes ); } finally { cOut.close(); }

      } catch( PGPException ex ) {
         throw new CryptoException( ex );
      } finally {
         try { in.close();                    } catch( Exception e ) {/*ok*/}
         try { encryptedOutpStrm.close();     } catch( Exception e ) {/*ok*/}
         try { encryptedOutputStream.close(); } catch( Exception e ) {/*ok*/}
         try { publicKeyInputStream.close();  } catch( Exception e ) {/*ok*/}
      }
   }

   private static byte[] compressFile( int algorithm, InputStream in, String inFilename, long inByteLength, Date inDataLastModified ) throws IOException
   {
      ByteArrayOutputStream bOut = new ByteArrayOutputStream();
      PGPCompressedDataGenerator comData = new PGPCompressedDataGenerator( algorithm );
      try {
         writeFileToLiteralData( comData.open( bOut ), PGPLiteralData.BINARY, in, inFilename, inByteLength, inDataLastModified );
      } finally {
         comData.close();
      }
      return bOut.toByteArray();
   }

   private static void writeFileToLiteralData( OutputStream out, char fileType,
         InputStream in, String inFilename, long inByteLength, Date inDataLastModified ) throws IOException
   {
      PGPLiteralDataGenerator lData = new PGPLiteralDataGenerator();
      OutputStream pOut = lData.open( out, fileType, inFilename, inByteLength, inDataLastModified );
      pipeFileContents( in, pOut, 4096 );
   }

   private static void pipeFileContents( InputStream in, OutputStream pOut, int bufSize ) throws IOException
   {
      try {
         byte[] buf = new byte[bufSize];
         int len;
         while( (len = in.read( buf )) > 0 ) { pOut.write( buf, 0, len ); }
      } finally {
         try { pOut.close(); } catch( Exception e ) {/*ok*/}
         try { in.close();   } catch( Exception e ) {/*ok*/}
      }
   }

   private static PGPPrivateKey findSecretKey( PGPSecretKeyRingCollection pgpSec, long keyID, char[] pass ) throws PGPException
   {
      PGPSecretKey pgpSecKey = pgpSec.getSecretKey( keyID );
      if( pgpSecKey == null ) { return null; }
      return pgpSecKey.extractPrivateKey( new JcePBESecretKeyDecryptorBuilder( new JcaPGPDigestCalculatorProviderBuilder()
                          .setProvider( new BouncyCastleProvider() ).build() ).setProvider( new BouncyCastleProvider() ).build( pass ) );
   }

   private static PGPPublicKey readPublicKey( InputStream input ) throws IOException, PGPException
   {
      PGPPublicKeyRingCollection pgpPub = new PGPPublicKeyRingCollection( PGPUtil.getDecoderStream( input ) );

      // we just loop through the collection till we find a key suitable for encryption,
      // in the real world you would probably want to be a bit smarter about this.
      Iterator<?> keyRingIter = pgpPub.getKeyRings();
      while( keyRingIter.hasNext() ) {
         PGPPublicKeyRing keyRing = (PGPPublicKeyRing) keyRingIter.next();
         Iterator<?> keyIter = keyRing.getPublicKeys();
         while( keyIter.hasNext() ) {
            PGPPublicKey key = (PGPPublicKey) keyIter.next();
            if( key.isEncryptionKey() ) { return key; }
         }
      }
      throw new CryptoException( "Can't find encryption key in key ring." );
   }

   /** RSA-Schluesselpaar erzeugen.<br>
    *  Die Streams secretKeyOut und publicKeyOut werden mit close() geschlossen. */
   private static void exportKeyPair( OutputStream secretKeyOut, OutputStream publicKeyOut, KeyPair keyPair,
         String identity, char[] passPhrase, boolean armor ) throws IOException, PGPException
   {
      OutputStream secretOut = secretKeyOut;
      OutputStream publicOut = publicKeyOut;
      PGPSecretKey secretKey = null;

      try {
         try {
            JcaPGPKeyConverter  jcaPGPKeyConverter = new JcaPGPKeyConverter();
            PGPPublicKey        pgpPublicKey  = jcaPGPKeyConverter.getPGPPublicKey( PUBLIC_KEY_ALGORITHM_TAG, keyPair.getPublic(), new Date() );
            PGPPrivateKey       pgpPrivateKey = jcaPGPKeyConverter.getPGPPrivateKey( pgpPublicKey, keyPair.getPrivate() );
            PGPKeyPair          pgpKeyPair    = new PGPKeyPair( pgpPublicKey, pgpPrivateKey );
            PGPDigestCalculator sha1Calc      = new JcaPGPDigestCalculatorProviderBuilder().build().get(HashAlgorithmTags.SHA1);
            secretKey = new PGPSecretKey( PGPSignature.DEFAULT_CERTIFICATION, pgpKeyPair, identity, sha1Calc, null, null,
                  new JcaPGPContentSignerBuilder( pgpKeyPair.getPublicKey().getAlgorithm(), HashAlgorithmTags.SHA1 ),
                  new JcePBESecretKeyEncryptorBuilder( SymmetricKeyAlgorithmTags.CAST5, sha1Calc ).setProvider( new BouncyCastleProvider() ).build( passPhrase ) );

            if( armor ) {
               secretOut = new ArmoredOutputStream( secretOut );
            }
            secretKey.encode( secretOut );

         } finally {
            secretOut.close();
         }

         if( armor ) {
            publicOut = new ArmoredOutputStream( publicOut );
         }
         PGPPublicKey key = secretKey.getPublicKey();
         key.encode( publicOut );

      } finally {
         publicOut.close();
      }
   }
}
