package de.meinefirma.meinprojekt.crypto;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.stream.EventFilter;
import javax.xml.stream.FactoryConfigurationError;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;

import org.apache.ftpserver.FtpServer;
import org.apache.ftpserver.ftplet.FtpException;
import org.junit.AfterClass;
import org.junit.Assert;
import org.junit.BeforeClass;
import org.junit.Test;

import de.meinefirma.meinprojekt.ftp.FtpFileWrapper;
import de.meinefirma.meinprojekt.ftp.FtpTestUtil;

/**
 * Teste das verschluesselte Schreiben von mit JAXB und StAX erzeugten XML-Daten auf FTP sowie
 * das Lesen von FTP, das Entschluesseln und das Umwandeln mit StAX und JAXB in Java-Objekte.<br>
 * Verwendet wird:<br>
 *   JAXB:         {@link javax.xml.bind.Marshaller }<br>
 *   StAX:         {@link javax.xml.stream.XMLStreamWriter }<br>
 *   BouncyCastle: {@link de.meinefirma.meinprojekt.crypto.OpenPgpBouncyCastleStreamUtil }<br>
 *   FTP:          {@link de.meinefirma.meinprojekt.ftp.FtpFileWrapper }<br>
 */
public class JaxbStaxCryptoFtpTest
{
   private static final int    FTP_PORT        = 2121;
   private static final String FTP_HOST        = "localhost";
   private static final String FTP_HOME_DIR    = "target/FtpHome";
   private static final String READ_USER_NAME  = "ReadUserName";
   private static final String READ_USER_PWD   = "ReadUserPwd";
   private static final String WRITE_USER_NAME = "WriteUserName";
   private static final String WRITE_USER_PWD  = "WriteUserPwd";
   private static final String KEYIDNAME       = "MeinName (BC-Test) <ich@myserver.de>";
   private static final String PASSPHRASE      = "Meine geheime BC-Passphrase 42";
   private static final String MEIN_TEXT       = "Blablupp \u00E4\u00F6\u00FC\u00DF\u20AC"; // Umlaute und Eurozeichen
   private static FtpServer ftpServer;
   private static byte[] privateKeyBA;
   private static byte[] publicKeyBA;

   @BeforeClass
   public static void start() throws FtpException, IOException, NoSuchAlgorithmException
   {
      // Starte FTP-Server:
      ftpServer = FtpTestUtil.createFtpServer( FTP_PORT, FTP_HOME_DIR,
            READ_USER_NAME, READ_USER_PWD, WRITE_USER_NAME, WRITE_USER_PWD );
      ftpServer.start();

      // Verschluesselung mit dem Bouncy-Castle-Provider vorbereiten:
      OpenPgpBouncyCastleStreamUtil.addBouncyCastleProvider();

      // Asymmetrisches Private/Public-Key-Paar generieren:
      ByteArrayOutputStream privateKeyOutputStream = new ByteArrayOutputStream();
      ByteArrayOutputStream publicKeyOutputStream  = new ByteArrayOutputStream();
      OpenPgpBouncyCastleStreamUtil.generateKeyPairStreams( privateKeyOutputStream, publicKeyOutputStream,
            KEYIDNAME, PASSPHRASE, true );
      privateKeyBA = privateKeyOutputStream.toByteArray();
      publicKeyBA  = publicKeyOutputStream.toByteArray();
   }

   @AfterClass
   public static void stopp()
   {
      ftpServer.stop();
      ftpServer = null;
   }

   @Test
   public void testJaxbStaxEncryptFtp() throws Exception
   {
      final String charEnc             = "UTF-8";
      final String remoteEncryptedFile = "MeineEncryptedDatei.bpg";
      final String remoteTempFile      = "MeineTempDatei$$";
      final String uri                 = "ftp://" + WRITE_USER_NAME + ":" + WRITE_USER_PWD + "@" + FTP_HOST + ":" + FTP_PORT;
      FtpFileWrapper ffwrap1           = FtpFileWrapper.createFtpFileWrapperFromUri( uri );
      FtpFileWrapper ffwrap2           = FtpFileWrapper.createFtpFileWrapperFromUri( uri );

      // Remote-Dateien loeschen, falls durch frueheren Testlauf vorhanden:
      ffwrap1.deleteFile( remoteTempFile );
      ffwrap1.deleteFile( remoteEncryptedFile );
      Assert.assertFalse( "delete", ffwrap1.existsFile( remoteTempFile ) );
      Assert.assertFalse( "delete", ffwrap1.existsFile( remoteEncryptedFile ) );

      // Speichere mit JAXB und StAX produzierte Daten auf FTP-Server:
      OutputStream osTemp = ffwrap1.storeFile( remoteTempFile );
      try {
         schreibeJaxbStax( osTemp, charEnc );
      } finally {
         osTemp.close();
         Assert.assertTrue( ffwrap1.completePendingCommand() );
      }

      // Lies von FTP und speichere auf FTP verschluesselt
      // (Achtung: ffwrap1.getLength() muss unbedingt vor ffwrap1.retrieveBinaryFile() aufgerufen werden!):
      long len = ffwrap1.getLength( remoteTempFile );
      InputStream  isTemp = ffwrap1.retrieveBinaryFile( remoteTempFile );
      OutputStream osEncr = ffwrap2.storeFile( remoteEncryptedFile );
      try {
         OpenPgpBouncyCastleStreamUtil.encryptStream( osEncr, isTemp, remoteTempFile, len,
               new Date(), new ByteArrayInputStream( publicKeyBA ), false, true );
      } finally {
         osEncr.close();
         isTemp.close();
         Assert.assertTrue( ffwrap2.completePendingCommand() );
         Assert.assertTrue( ffwrap1.completePendingCommand() );
      }
      ffwrap2.logout();

      // Vorlaeufige Ueberpruefung der Ergebnisdatei auf dem FTP-Server:
      Assert.assertTrue( ffwrap1.existsFile( remoteEncryptedFile ) );
      Assert.assertTrue( ffwrap1.getLength( remoteEncryptedFile ) > 400 );

      // Lies vom FTP-Server, und versuche Entschluesselung mit falscher Passphrase:
      InputStream inFtp = ffwrap1.retrieveBinaryFile( remoteEncryptedFile );
      InputStream inDecrypted = null;
      try {
         OpenPgpBouncyCastleStreamUtil bouncyCastleUtil = new OpenPgpBouncyCastleStreamUtil();
         inDecrypted = bouncyCastleUtil.decryptStream( inFtp, 
               new ByteArrayInputStream( privateKeyBA ), 
               (PASSPHRASE + "x").toCharArray() );
         Assert.fail();
      } catch( CryptoException ce ) {
         Assert.assertTrue( ce.toString().contains( "org.bouncycastle.openpgp.PGPException:" ) );
      } finally {
         if( inDecrypted != null ) { inDecrypted.close(); }
         inFtp.close();
         ffwrap1.completePendingCommand();
      }

      // Lies vom FTP-Server, entschluessle und wandel per StAX und JAXB zu Java-Objekten:
      inFtp = ffwrap1.retrieveBinaryFile( remoteEncryptedFile );
      inDecrypted = null;
      try {
         OpenPgpBouncyCastleStreamUtil bouncyCastleUtil = new OpenPgpBouncyCastleStreamUtil();
         inDecrypted = bouncyCastleUtil.decryptStream( inFtp, 
               new ByteArrayInputStream( privateKeyBA ), 
               PASSPHRASE.toCharArray() );
         liesStaxJaxb( inDecrypted, charEnc );
         Assert.assertTrue( bouncyCastleUtil.verifyIntegrityProtected() );
      } finally {
         if( inDecrypted != null ) { inDecrypted.close(); }
         inFtp.close();
         Assert.assertTrue( ffwrap1.completePendingCommand() );
      }
   }

   static void liesStaxJaxb( InputStream in, String charEnc ) throws Exception
   {
      EventFilter startElementFilter = new EventFilter() {
         @Override public boolean accept( XMLEvent event ) {
            return event.isStartElement();
         }
      };
      try {
         XMLInputFactory staxFactory  = XMLInputFactory.newInstance();
         XMLEventReader  staxReader   = staxFactory.createXMLEventReader( new InputStreamReader( in, charEnc ) );
         XMLEventReader  staxFiltRd   = staxFactory.createFilteredReader( staxReader, startElementFilter );
         JAXBContext     jaxbContext  = JAXBContext.newInstance( MeineDatenKlasse.class );
         Unmarshaller    unmarshaller = jaxbContext.createUnmarshaller();
         StartElement    rootElement  = staxFiltRd.nextEvent().asStartElement();
         Assert.assertEquals( "<MeinRootElement>", rootElement.toString() );
         while( staxFiltRd.peek() != null ) {
            Object element = unmarshaller.unmarshal( staxReader );
            Assert.assertTrue( element instanceof MeineDatenKlasse );
            MeineDatenKlasse mdk = (MeineDatenKlasse) element;
            Assert.assertEquals( MEIN_TEXT, mdk.s );
         }
      } finally {
         in.close();
      }
   }

   static void schreibeJaxbStax( OutputStream os, String charEnc ) throws XMLStreamException, FactoryConfigurationError, JAXBException, IOException
   {
      try {
         XMLStreamWriter xsw = XMLOutputFactory.newInstance().createXMLStreamWriter( new OutputStreamWriter( os, charEnc ) );
         // xsw = new IndentingXMLStreamWriter( xsw );
         try {
            Marshaller marshaller = JAXBContext.newInstance( MeinRootElement.class ).createMarshaller();
            marshaller.setProperty( Marshaller.JAXB_ENCODING, charEnc );
            marshaller.setProperty( Marshaller.JAXB_FRAGMENT, Boolean.TRUE );

            xsw.writeStartDocument( charEnc, null );
            xsw.writeCharacters(   "\n" );
            xsw.writeStartElement( "MeinRootElement" );
            xsw.writeCharacters(   "\n" );

            MeineDatenKlasseDatenLieferant mdkdl = new MeineDatenKlasseDatenLieferant();
            while( mdkdl.hasNext() ) { marshaller.marshal( mdkdl.next(), xsw ); xsw.writeCharacters( "\n" ); }

            xsw.writeEndElement();
            xsw.writeEndDocument();
            xsw.writeCharacters( "\n" );
            xsw.flush();
         } finally {
            xsw.close();
         }
      } finally {
         os.close();
      }
   }

   static class MeineDatenKlasseDatenLieferant implements Iterator<MeineDatenKlasse> {
      private int idx = 0;
      @Override public void remove() { throw new UnsupportedOperationException(); }
      @Override public boolean hasNext() { return this.idx < 10; }
      @Override public MeineDatenKlasse next() {
         MeineDatenKlasse mdk = new MeineDatenKlasse();
         mdk.i = 4711 + this.idx++;
         mdk.s = MEIN_TEXT;
         return mdk;
      }
   }
}

@XmlRootElement
class MeineDatenKlasse
{
   public int    i;
   public String s;
}

@XmlRootElement
class MeinRootElement
{
   public List<MeineDatenKlasse> mdks = new ArrayList<MeineDatenKlasse>();
}
