package de.meinefirma.meinprojekt.crypto;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.security.GeneralSecurityException;
import java.util.Date;

import org.apache.ftpserver.FtpServer;
import org.apache.ftpserver.ftplet.FtpException;
import org.junit.AfterClass;
import org.junit.Assert;
import org.junit.BeforeClass;
import org.junit.Test;

import de.meinefirma.meinprojekt.ftp.FtpFileWrapper;
import de.meinefirma.meinprojekt.ftp.FtpTestUtil;

public class OpenPgpBouncyCastleStreamUtilFtpTest
{
   private static final int    FTP_PORT        = 2121;
   private static final String FTP_HOST        = "localhost";
   private static final String FTP_HOME_DIR    = "target/FtpHome";
   private static final String READ_USER_NAME  = "ReadUserName";
   private static final String READ_USER_PWD   = "ReadUserPwd";
   private static final String WRITE_USER_NAME = "WriteUserName";
   private static final String WRITE_USER_PWD  = "WriteUserPwd";
   private static FtpServer ftpServer;

   @BeforeClass
   public static void startFtpServer() throws FtpException, IOException
   {
      ftpServer = FtpTestUtil.createFtpServer( FTP_PORT, FTP_HOME_DIR,
            READ_USER_NAME, READ_USER_PWD, WRITE_USER_NAME, WRITE_USER_PWD );
      ftpServer.start();
   }

   @AfterClass
   public static void stoppFtpServer()
   {
      ftpServer.stop();
      ftpServer = null;
   }

   /**
    * Teste Fileserver-Variante mit FS-URI
    */
   @Test
   public void testFsMitUri() throws GeneralSecurityException, IOException
   {
      final String uri = "target/FileserverHome";
      FtpFileWrapper ffwrap = FtpFileWrapper.createFtpFileWrapperFromUri( uri );
      Assert.assertEquals( "{FtpUri=target/FileserverHome}", ffwrap.getPropsAsString() );
      testBcFtp( ffwrap );
   }

   /**
    * Teste FTP-Variante mit FTP-URI
    */
   @Test
   public void testFtpMitUri() throws GeneralSecurityException, IOException
   {
      final String uri = "ftp://" + WRITE_USER_NAME + ":" + WRITE_USER_PWD + "@" + FTP_HOST + ":" + FTP_PORT;
      FtpFileWrapper ffwrap = FtpFileWrapper.createFtpFileWrapperFromUri( uri );
      Assert.assertEquals( "{FtpUri=ftp://WriteUserName:...@localhost:2121}", ffwrap.getPropsAsString() );
      testBcFtp( ffwrap );
   }

   /** Interne gemeinsame Testmethode sowohl fuer FTP- als auch FS-Betrieb */
   protected static void testBcFtp( FtpFileWrapper ffwrap ) throws GeneralSecurityException, IOException
   {
      testBcFtpOhneException( ffwrap );
      testBcFtpMitExceptionFalschePassphrase( ffwrap );
      testBcFtpOhneException( ffwrap );
   }

   protected static void testBcFtpOhneException( FtpFileWrapper ffwrap ) throws GeneralSecurityException, IOException
   {
      String meinText   = "Blablupp \u00E4\u00F6\u00FC\u00DF\u20AC"; // Umlaute und Eurozeichen
      String charEnc    = "UTF-8";
      String keyIdName  = "MeinName (BC-Test) <ich@myserver.de>";
      String passphrase = "Meine geheime BC-Passphrase 42";
      String remoteFile = "MeineEncryptedDatei.bpg";

      // Verschluesselung mit dem Bouncy-Castle-Provider vorbereiten:
      OpenPgpBouncyCastleStreamUtil.addBouncyCastleProvider();

      // Asymmetrisches Private/Public-Key-Paar generieren:
      ByteArrayOutputStream privateKeyOutputStream = new ByteArrayOutputStream();
      ByteArrayOutputStream publicKeyOutputStream  = new ByteArrayOutputStream();
      OpenPgpBouncyCastleStreamUtil.generateKeyPairStreams( privateKeyOutputStream, publicKeyOutputStream,
            keyIdName, passphrase, true );
      byte[] privateKeyBA = privateKeyOutputStream.toByteArray();
      byte[] publicKeyBA  = publicKeyOutputStream.toByteArray();
      Assert.assertTrue( privateKeyBA.length > 0 );
      Assert.assertTrue( publicKeyBA.length  > 0 );

      // Remote-Datei loeschen, falls durch frueheren Testlauf vorhanden:
      ffwrap.deleteFile( remoteFile );
      Assert.assertFalse( "delete", ffwrap.existsFile( remoteFile ) );

      // Verschluesseln und per FTP speichern:
      byte[] meinTextBA = meinText.getBytes( charEnc );
      OutputStream os = ffwrap.storeFile( remoteFile );
      try {
         OpenPgpBouncyCastleStreamUtil.encryptStream( os, new ByteArrayInputStream( meinTextBA ),
               "MeinText", meinTextBA.length, new Date(), new ByteArrayInputStream( publicKeyBA ), false, true );
      } finally {
         os.close();
         Assert.assertTrue( "completePendingCommand", ffwrap.completePendingCommand() );
      }
      Assert.assertTrue( "exists", ffwrap.existsFile( remoteFile ) );

      // Von FTP downloaden und entschluesseln mit OutputStream-decrypt-Variante:
      ByteArrayOutputStream decryptedOut = new ByteArrayOutputStream();
      InputStream is1 = ffwrap.retrieveBinaryFile( remoteFile );
      try {
         Assert.assertTrue( OpenPgpBouncyCastleStreamUtil.decryptStream( decryptedOut, is1,
               new ByteArrayInputStream( privateKeyBA ), passphrase.toCharArray() ) );
      } finally {
         is1.close();
         Assert.assertTrue( "completePendingCommand", ffwrap.completePendingCommand() );
      }
      String decryptedText1 = decryptedOut.toString( charEnc );
      Assert.assertEquals( meinText, decryptedText1 );

      // Von FTP downloaden und entschluesseln mit InputStream-decrypt-Variante:
      InputStream is2 = ffwrap.retrieveBinaryFile( remoteFile );
      OpenPgpBouncyCastleStreamUtil opbcsu = new OpenPgpBouncyCastleStreamUtil();
      try {
         InputStream is = opbcsu.decryptStream( is2, new ByteArrayInputStream( privateKeyBA ), passphrase.toCharArray() );
         BufferedReader in = new BufferedReader( new InputStreamReader( is, charEnc ) );
         String decryptedText2 = in.readLine();
         in.close();
         is.close();
         Assert.assertEquals( meinText, decryptedText2 );
      } finally {
         is2.close();
         Assert.assertTrue( "completePendingCommand", ffwrap.completePendingCommand() );
      }
      Assert.assertTrue( "verifyIntegrityProtected", opbcsu.verifyIntegrityProtected() );
   }

   protected static void testBcFtpMitExceptionFalschePassphrase( FtpFileWrapper ffwrap ) throws GeneralSecurityException, IOException
   {
      String meinText   = "Blablupp \u00E4\u00F6\u00FC\u00DF\u20AC"; // Umlaute und Eurozeichen
      String charEnc    = "UTF-8";
      String keyIdName  = "MeinName (BC-Test) <ich@myserver.de>";
      String passphrase = "Meine geheime BC-Passphrase 42";
      String remoteFile = "MeineEncryptedDatei.bpg";

      // Verschluesselung mit dem Bouncy-Castle-Provider vorbereiten:
      OpenPgpBouncyCastleStreamUtil.addBouncyCastleProvider();

      // Asymmetrisches Private/Public-Key-Paar generieren:
      ByteArrayOutputStream privateKeyOutputStream = new ByteArrayOutputStream();
      ByteArrayOutputStream publicKeyOutputStream  = new ByteArrayOutputStream();
      OpenPgpBouncyCastleStreamUtil.generateKeyPairStreams( privateKeyOutputStream, publicKeyOutputStream,
            keyIdName, passphrase, true );
      byte[] privateKeyBA = privateKeyOutputStream.toByteArray();
      byte[] publicKeyBA  = publicKeyOutputStream.toByteArray();
      Assert.assertTrue( privateKeyBA.length > 0 );
      Assert.assertTrue( publicKeyBA.length  > 0 );

      // Remote-Datei loeschen, falls durch frueheren Testlauf vorhanden:
      ffwrap.deleteFile( remoteFile );
      Assert.assertFalse( "delete", ffwrap.existsFile( remoteFile ) );

      // Verschluesseln und per FTP speichern:
      byte[] meinTextBA = meinText.getBytes( charEnc );
      OutputStream os = ffwrap.storeFile( remoteFile );
      try {
         OpenPgpBouncyCastleStreamUtil.encryptStream( os, new ByteArrayInputStream( meinTextBA ),
               "MeinText", meinTextBA.length, new Date(), new ByteArrayInputStream( publicKeyBA ), false, true );
      } finally {
         os.close();
         Assert.assertTrue( "completePendingCommand", ffwrap.completePendingCommand() );
      }
      Assert.assertTrue( "exists", ffwrap.existsFile( remoteFile ) );

      // Von FTP downloaden und entschluesseln mit OutputStream-decrypt-Variante:
      ByteArrayOutputStream decryptedOut = new ByteArrayOutputStream();
      InputStream is1 = ffwrap.retrieveBinaryFile( remoteFile );
      try {
         // Falsche Passphrase muss Exception werfen:
         OpenPgpBouncyCastleStreamUtil.decryptStream( decryptedOut, is1,
               new ByteArrayInputStream( privateKeyBA ), (passphrase + "x").toCharArray() );
         Assert.fail( "Falsche Passphrase." );
      } catch( CryptoException ce ) {
         Assert.assertTrue( ce.toString().contains( "org.bouncycastle.openpgp.PGPException:" ) );
      } finally {
         is1.close();
         Assert.assertTrue( "completePendingCommand", ffwrap.completePendingCommand() );
      }
      String decryptedText = decryptedOut.toString( charEnc );
      Assert.assertEquals( "", decryptedText );

      // Von FTP downloaden und entschluesseln mit InputStream-decrypt-Variante:
      InputStream is2 = ffwrap.retrieveBinaryFile( remoteFile );
      OpenPgpBouncyCastleStreamUtil opbcsu = new OpenPgpBouncyCastleStreamUtil();
      try {
         // Falsche Passphrase muss Exception werfen:
         opbcsu.decryptStream( is2, new ByteArrayInputStream( privateKeyBA ), (passphrase + "x").toCharArray() );
         Assert.fail( "Falsche Passphrase." );
      } catch( CryptoException ce ) {
         Assert.assertTrue( ce.toString().contains( "org.bouncycastle.openpgp.PGPException:" ) );
      } finally {
         is2.close();
         Assert.assertTrue( "completePendingCommand", ffwrap.completePendingCommand() );
      }
   }
}
