package de.meinefirma.meinprojekt.crypto;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.StringWriter;
import java.io.Writer;
import java.security.GeneralSecurityException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.junit.Assert;
import org.junit.Test;

public class OpenPgpBouncyCastleStreamUtilTest
{
   // Test ohne Dateien, nur mit Streams
   @Test
   public void testBcStreamByteArray() throws GeneralSecurityException, IOException
   {
      String meinText   = "Blablupp \u00E4\u00F6\u00FC\u00DF\u20AC"; // Umlaute und Eurozeichen
      String charEnc    = "UTF-8";
      String keyIdName  = "MeinName (BC-Test) <ich@myserver.de>";
      String passphrase = "Meine geheime BC-Passphrase 42";

      // Verschluesselung mit dem Bouncy-Castle-Provider vorbereiten:
      OpenPgpBouncyCastleStreamUtil.addBouncyCastleProvider();

      // Asymmetrisches Private/Public-Key-Paar generieren:
      ByteArrayOutputStream privateKeyOutputStream = new ByteArrayOutputStream();
      ByteArrayOutputStream publicKeyOutputStream  = new ByteArrayOutputStream();
      OpenPgpBouncyCastleStreamUtil.generateKeyPairStreams( privateKeyOutputStream, publicKeyOutputStream,
            keyIdName, passphrase, true );
      byte[] privateKeyBA = privateKeyOutputStream.toByteArray();
      byte[] publicKeyBA  = publicKeyOutputStream.toByteArray();
      Assert.assertTrue( privateKeyBA.length > 0 );
      Assert.assertTrue( publicKeyBA.length  > 0 );

      // Verschluesseln:
      byte[] meinTextBA = meinText.getBytes( charEnc );
      ByteArrayOutputStream encryptedOut = new ByteArrayOutputStream();
      OpenPgpBouncyCastleStreamUtil.encryptStream( encryptedOut, new ByteArrayInputStream( meinTextBA ),
            "MeinText", meinTextBA.length, new Date(), new ByteArrayInputStream( publicKeyBA ), false, true );
      byte[] encryptedBA = encryptedOut.toByteArray();
      Assert.assertTrue( encryptedBA.length > 0 );

      // In einen OutputStream entschluesseln:
      ByteArrayOutputStream decryptedOut = new ByteArrayOutputStream();
      OpenPgpBouncyCastleStreamUtil.decryptStream( decryptedOut, new ByteArrayInputStream( encryptedBA ),
            new ByteArrayInputStream( privateKeyBA ), passphrase.toCharArray() );
      String decryptedText = decryptedOut.toString( charEnc );
      Assert.assertEquals( meinText, decryptedText );
      
      // Zu einem InputStream entschluesseln:
      OpenPgpBouncyCastleStreamUtil bcUtil = new OpenPgpBouncyCastleStreamUtil();
      InputStream in = bcUtil.decryptStream( new ByteArrayInputStream( encryptedBA ),
            new ByteArrayInputStream( privateKeyBA ), passphrase.toCharArray() );                
      decryptedText = getInputStreamAsString( in, charEnc );
      in.close();
      Assert.assertEquals( meinText, decryptedText );
      Assert.assertTrue( bcUtil.verifyIntegrityProtected() );
   }

   private static String getInputStreamAsString(InputStream in, String charEnc) throws IOException
   {
      Writer writer = new StringWriter();             
      char[] buffer = new char[1024];                        
      Reader reader = new BufferedReader(                     
      new InputStreamReader(in, charEnc));               
      int n;             
      while ((n = reader.read(buffer)) != -1) {                  
         writer.write(buffer, 0, n);             
      }
      return writer.toString();
   }

   // Test mit Dateien
   @Test
   public void testBcStreamFile() throws GeneralSecurityException, IOException
   {
      String text           = "Blablupp \u00E4\u00F6\u00FC\u00DF\u20AC"; // Umlaute undEurozeichen
      String charEnc        = "UTF-8";
      String keyIdName      = "MeinName (BC-Test) <ich@myserver.de>";
      String passphrase     = "Meine geheime BC-Passphrase 42";
      String privateKeyFile = "target/PrivateKey.rsa";
      String publicKeyFile  = "target/PublicKey.rsa";
      String srcFile        = "target/MeineSrcDatei.txt";
      String dstFile        = "target/MeineDstDatei.txt";
      String encFile        = "target/MeineVerschluesselteDatei.bpg";

      OpenPgpBouncyCastleStreamUtil.addBouncyCastleProvider();

      (new File( srcFile )).delete();
      (new File( dstFile )).delete();
      (new File( encFile )).delete();
      BufferedWriter out = new BufferedWriter( new OutputStreamWriter( new FileOutputStream( srcFile ), charEnc ) );
      try { out.write( text ); } finally { out.close(); }

      OpenPgpBouncyCastleStreamUtil.generateKeyPairFiles( privateKeyFile, publicKeyFile, keyIdName, passphrase, true );
      Assert.assertTrue( new File( privateKeyFile ).exists() );
      Assert.assertTrue( new File( publicKeyFile ).exists() );

      OpenPgpBouncyCastleStreamUtil.encryptFile( encFile, srcFile, publicKeyFile, false, true );
      Assert.assertTrue( new File( encFile ).exists() );

      Assert.assertFalse( new File( dstFile ).exists() );
      OpenPgpBouncyCastleStreamUtil.decryptFile( dstFile, encFile, privateKeyFile, passphrase );
      Assert.assertTrue( new File( dstFile ).exists() );

      List<String> ss = readTextFile( dstFile, charEnc );
      Assert.assertEquals( 1, ss.size() );
      Assert.assertEquals( text, ss.get( 0 ) );
   }

   // Test mit XML-Dateien
   @Test
   public void testBcStreamXmlFile() throws GeneralSecurityException, IOException
   {
      String charEnc        = "UTF-8";
      String keyIdName      = "MeinName (BC-Test) <ich@myserver.de>";
      String passphrase     = "Meine geheime BC-Passphrase 42";
      String privateKeyFile = "target/PrivateXmlKey.rsa";
      String publicKeyFile  = "target/PublicXmlKey.rsa";
      String srcFile        = "src/test/resources/MeineXmlSrcDatei.xml";
      String encFile        = "target/MeineXmlVerschluesselteDatei.xml.bpg";

      OpenPgpBouncyCastleStreamUtil.addBouncyCastleProvider();

      new File( encFile ).delete();

      OpenPgpBouncyCastleStreamUtil.generateKeyPairFiles( privateKeyFile, publicKeyFile, keyIdName, passphrase, true );
      Assert.assertTrue( new File( privateKeyFile ).exists() );
      Assert.assertTrue( new File( publicKeyFile ).exists() );

      // Xml-Datei verschluesseln
      OpenPgpBouncyCastleStreamUtil.encryptFile( encFile, srcFile, publicKeyFile, false, true );
      Assert.assertTrue( new File( encFile ).exists() );

      OpenPgpBouncyCastleStreamUtil bcUtil = new OpenPgpBouncyCastleStreamUtil();
      InputStream pk = new FileInputStream( privateKeyFile );
      InputStream is = new FileInputStream( encFile );
      InputStream in = bcUtil.decryptStream( is, pk, passphrase.toCharArray() );                
      String decryptedText = getInputStreamAsString( in, charEnc );
      in.close();
      is.close();
      pk.close();
      Assert.assertNotNull( decryptedText );
      Assert.assertTrue( bcUtil.verifyIntegrityProtected() ); 
   }

   static List<String> readTextFile( String file, String charEnc ) throws IOException
   {
      List<String>   ss = new ArrayList<String>();
      BufferedReader in = new BufferedReader( new InputStreamReader( new FileInputStream( file ), charEnc ) );
      try {
         String line;
         while( (line = in.readLine()) != null ) { ss.add( line ); }
      } finally {
         in.close();
      }
      return ss;
   }
}
